@@ -1,10 +1,12 @@
-/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/data.c,v 1.42 2009/01/15 11:52:55 petere Exp $ */
+/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/data.c,v 1.53 2010/05/25 17:28:20 meskes Exp $ */
 
 #define POSTGRES_ECPG_INTERNAL
 #include "postgres_fe.h"
 
 #include <stdlib.h>
 #include <string.h>
+#include <float.h>
+#include <math.h>
 
 #include "ecpgtype.h"
 #include "ecpglib.h"
@@ -16,6 +18,33 @@
 #include "pgtypes_timestamp.h"
 #include "pgtypes_interval.h"
 
+/* returns true if character c is a delimiter for the given array type */
+static bool
+array_delimiter(enum ARRAY_TYPE isarray, char c)
+{
+	if (isarray == ECPG_ARRAY_ARRAY && c == ',')
+		return true;
+
+	if (isarray == ECPG_ARRAY_VECTOR && c == ' ')
+		return true;
+
+	return false;
+}
+
+/* returns true if character c marks the boundary for the given array type */
+static bool
+array_boundary(enum ARRAY_TYPE isarray, char c)
+{
+	if (isarray == ECPG_ARRAY_ARRAY && c == '}')
+		return true;
+
+	if (isarray == ECPG_ARRAY_VECTOR && c == '\0')
+		return true;
+
+	return false;
+}
+
+/* returns true if some garbage is found at the end of the scanned string */
 static bool
 garbage_left(enum ARRAY_TYPE isarray, char *scan_length, enum COMPAT_MODE compat)
 {
@@ -23,17 +52,69 @@ garbage_left(enum ARRAY_TYPE isarray, char *scan_length, enum COMPAT_MODE compat
 	 * INFORMIX allows for selecting a numeric into an int, the result is
 	 * truncated
 	 */
-	if (isarray == ECPG_ARRAY_NONE && INFORMIX_MODE(compat) && *scan_length == '.')
-		return false;
+	if (isarray == ECPG_ARRAY_NONE)
+	{
+		if (INFORMIX_MODE(compat) && *scan_length == '.')
+			return false;
 
-	if (isarray == ECPG_ARRAY_ARRAY && *scan_length != ',' && *scan_length != '}')
+		if (*scan_length != ' ' && *scan_length != '\0')
+			return true;
+	}
+	else if (ECPG_IS_ARRAY(isarray) && !array_delimiter(isarray, *scan_length) && !array_boundary(isarray, *scan_length))
 		return true;
 
-	if (isarray == ECPG_ARRAY_VECTOR && *scan_length != ' ' && *scan_length != '\0')
-		return true;
+	return false;
+}
 
-	if (isarray == ECPG_ARRAY_NONE && *scan_length != ' ' && *scan_length != '\0')
+/* stolen code from src/backend/utils/adt/float.c */
+#if defined(WIN32) && !defined(NAN)
+static const uint32 nan[2] = {0xffffffff, 0x7fffffff};
+
+#define NAN (*(const double *) nan)
+#endif
+
+static double
+get_float8_infinity(void)
+{
+#ifdef INFINITY
+	return (double) INFINITY;
+#else
+	return (double) (HUGE_VAL * HUGE_VAL);
+#endif
+}
+
+static double
+get_float8_nan(void)
+{
+	/* (double) NAN doesn't work on some NetBSD/MIPS releases */
+#if defined(NAN) && !(defined(__NetBSD__) && defined(__mips__))
+	return (double) NAN;
+#else
+	return (double) (0.0 / 0.0);
+#endif
+}
+
+static bool
+check_special_value(char *ptr, double *retval, char **endptr)
+{
+	if (!pg_strncasecmp(ptr, "NaN", 3))
+	{
+		*retval = get_float8_nan();
+		*endptr = ptr + 3;
 		return true;
+	}
+	else if (!pg_strncasecmp(ptr, "Infinity", 8))
+	{
+		*retval = get_float8_infinity();
+		*endptr = ptr + 8;
+		return true;
+	}
+	else if (!pg_strncasecmp(ptr, "-Infinity", 9))
+	{
+		*retval = -get_float8_infinity();
+		*endptr = ptr + 9;
+		return true;
+	}
 
 	return false;
 }
@@ -60,7 +141,18 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 	else
 		log_offset = offset;
 
-	ecpg_log("ecpg_get_data on line %d: RESULT: %s offset: %ld; array: %s\n", lineno, pval ? (binary ? "BINARY" : pval) : "EMPTY", log_offset, isarray ? "yes" : "no");
+	ecpg_log("ecpg_get_data on line %d: RESULT: %s offset: %ld; array: %s\n", lineno, pval ? (binary ? "BINARY" : pval) : "EMPTY", log_offset, ECPG_IS_ARRAY(isarray) ? "yes" : "no");
+
+	/* pval is a pointer to the value */
+	if (!pval)
+	{
+		/*
+		 * This should never happen because we already checked that we found
+		 * at least one tuple, but let's play it safe.
+		 */
+		ecpg_raise(lineno, ECPG_NOT_FOUND, ECPG_SQLSTATE_NO_DATA, NULL);
+		return (false);
+	}
 
 	/* We will have to decode the value */
 
@@ -85,12 +177,12 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 		case ECPGt_unsigned_long:
 			*((long *) (ind + ind_offset * act_tuple)) = value_for_indicator;
 			break;
-#ifdef HAVE_LONG_LONG_INT_64
+#ifdef HAVE_LONG_LONG_INT
 		case ECPGt_long_long:
 		case ECPGt_unsigned_long_long:
 			*((long long int *) (ind + ind_offset * act_tuple)) = value_for_indicator;
 			break;
-#endif   /* HAVE_LONG_LONG_INT_64 */
+#endif   /* HAVE_LONG_LONG_INT */
 		case ECPGt_NO_INDICATOR:
 			if (value_for_indicator == -1)
 			{
@@ -122,11 +214,10 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 	if (value_for_indicator == -1)
 		return (true);
 
-	/* pval is a pointer to the value */
 	/* let's check if it really is an array if it should be one */
 	if (isarray == ECPG_ARRAY_ARRAY)
 	{
-		if (!pval || *pval != '{')
+		if (*pval != '{')
 		{
 			ecpg_raise(lineno, ECPG_DATA_NOT_ARRAY,
 					   ECPG_SQLSTATE_DATATYPE_MISMATCH, NULL);
@@ -138,6 +229,7 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 			case ECPGt_char:
 			case ECPGt_unsigned_char:
 			case ECPGt_varchar:
+			case ECPGt_string:
 				break;
 
 			default:
@@ -150,47 +242,42 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 	{
 		if (binary)
 		{
-			if (pval)
+			if (varcharsize == 0 || varcharsize * offset >= size)
+				memcpy(var + offset * act_tuple, pval, size);
+			else
 			{
-				if (varcharsize == 0 || varcharsize * offset >= size)
-					memcpy((char *) ((long) var + offset * act_tuple),
-						   pval, size);
-				else
-				{
-					memcpy((char *) ((long) var + offset * act_tuple),
-						   pval, varcharsize * offset);
+				memcpy(var + offset * act_tuple, pval, varcharsize * offset);
 
-					if (varcharsize * offset < size)
+				if (varcharsize * offset < size)
+				{
+					/* truncation */
+					switch (ind_type)
 					{
-						/* truncation */
-						switch (ind_type)
-						{
-							case ECPGt_short:
-							case ECPGt_unsigned_short:
-								*((short *) (ind + ind_offset * act_tuple)) = size;
-								break;
-							case ECPGt_int:
-							case ECPGt_unsigned_int:
-								*((int *) (ind + ind_offset * act_tuple)) = size;
-								break;
-							case ECPGt_long:
-							case ECPGt_unsigned_long:
-								*((long *) (ind + ind_offset * act_tuple)) = size;
-								break;
-#ifdef HAVE_LONG_LONG_INT_64
-							case ECPGt_long_long:
-							case ECPGt_unsigned_long_long:
-								*((long long int *) (ind + ind_offset * act_tuple)) = size;
-								break;
-#endif   /* HAVE_LONG_LONG_INT_64 */
-							default:
-								break;
-						}
-						sqlca->sqlwarn[0] = sqlca->sqlwarn[1] = 'W';
+						case ECPGt_short:
+						case ECPGt_unsigned_short:
+							*((short *) (ind + ind_offset * act_tuple)) = size;
+							break;
+						case ECPGt_int:
+						case ECPGt_unsigned_int:
+							*((int *) (ind + ind_offset * act_tuple)) = size;
+							break;
+						case ECPGt_long:
+						case ECPGt_unsigned_long:
+							*((long *) (ind + ind_offset * act_tuple)) = size;
+							break;
+#ifdef HAVE_LONG_LONG_INT
+						case ECPGt_long_long:
+						case ECPGt_unsigned_long_long:
+							*((long long int *) (ind + ind_offset * act_tuple)) = size;
+							break;
+#endif   /* HAVE_LONG_LONG_INT */
+						default:
+							break;
 					}
+					sqlca->sqlwarn[0] = sqlca->sqlwarn[1] = 'W';
 				}
-				pval += size;
 			}
+			pval += size;
 		}
 		else
 		{
@@ -208,19 +295,14 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 				case ECPGt_short:
 				case ECPGt_int:
 				case ECPGt_long:
-					if (pval)
+					res = strtol(pval, &scan_length, 10);
+					if (garbage_left(isarray, scan_length, compat))
 					{
-						res = strtol(pval, &scan_length, 10);
-						if (garbage_left(isarray, scan_length, compat))
-						{
-							ecpg_raise(lineno, ECPG_INT_FORMAT,
-									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-							return (false);
-						}
-						pval = scan_length;
+						ecpg_raise(lineno, ECPG_INT_FORMAT,
+								   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+						return (false);
 					}
-					else
-						res = 0L;
+					pval = scan_length;
 
 					switch (type)
 					{
@@ -242,19 +324,14 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 				case ECPGt_unsigned_short:
 				case ECPGt_unsigned_int:
 				case ECPGt_unsigned_long:
-					if (pval)
+					ures = strtoul(pval, &scan_length, 10);
+					if (garbage_left(isarray, scan_length, compat))
 					{
-						ures = strtoul(pval, &scan_length, 10);
-						if (garbage_left(isarray, scan_length, compat))
-						{
-							ecpg_raise(lineno, ECPG_UINT_FORMAT,
-									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-							return (false);
-						}
-						pval = scan_length;
+						ecpg_raise(lineno, ECPG_UINT_FORMAT,
+								   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+						return (false);
 					}
-					else
-						ures = 0L;
+					pval = scan_length;
 
 					switch (type)
 					{
@@ -273,66 +350,52 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 					}
 					break;
 
-#ifdef HAVE_LONG_LONG_INT_64
+#ifdef HAVE_LONG_LONG_INT
 #ifdef HAVE_STRTOLL
 				case ECPGt_long_long:
-					if (pval)
+					*((long long int *) (var + offset * act_tuple)) = strtoll(pval, &scan_length, 10);
+					if (garbage_left(isarray, scan_length, compat))
 					{
-						*((long long int *) (var + offset * act_tuple)) = strtoll(pval, &scan_length, 10);
-						if (garbage_left(isarray, scan_length, compat))
-						{
-							ecpg_raise(lineno, ECPG_INT_FORMAT, ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-							return (false);
-						}
-						pval = scan_length;
+						ecpg_raise(lineno, ECPG_INT_FORMAT, ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+						return (false);
 					}
-					else
-						*((long long int *) (var + offset * act_tuple)) = (long long) 0;
+					pval = scan_length;
 
 					break;
 #endif   /* HAVE_STRTOLL */
 #ifdef HAVE_STRTOULL
 				case ECPGt_unsigned_long_long:
-					if (pval)
+					*((unsigned long long int *) (var + offset * act_tuple)) = strtoull(pval, &scan_length, 10);
+					if ((isarray && *scan_length != ',' && *scan_length != '}')
+						|| (!isarray && !(INFORMIX_MODE(compat) && *scan_length == '.') && *scan_length != '\0' && *scan_length != ' '))		/* Garbage left */
 					{
-						*((unsigned long long int *) (var + offset * act_tuple)) = strtoull(pval, &scan_length, 10);
-						if ((isarray && *scan_length != ',' && *scan_length != '}')
-							|| (!isarray && !(INFORMIX_MODE(compat) && *scan_length == '.') && *scan_length != '\0' && *scan_length != ' '))	/* Garbage left */
-						{
-							ecpg_raise(lineno, ECPG_UINT_FORMAT, ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-							return (false);
-						}
-						pval = scan_length;
+						ecpg_raise(lineno, ECPG_UINT_FORMAT, ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+						return (false);
 					}
-					else
-						*((unsigned long long int *) (var + offset * act_tuple)) = (long long) 0;
+					pval = scan_length;
 
 					break;
 #endif   /* HAVE_STRTOULL */
-#endif   /* HAVE_LONG_LONG_INT_64 */
+#endif   /* HAVE_LONG_LONG_INT */
 
 				case ECPGt_float:
 				case ECPGt_double:
-					if (pval)
+					if (isarray && *pval == '"')
+						pval++;
+
+					if (!check_special_value(pval, &dres, &scan_length))
+						dres = strtod(pval, &scan_length);
+
+					if (isarray && *scan_length == '"')
+						scan_length++;
+
+					if (garbage_left(isarray, scan_length, compat))
 					{
-						if (isarray && *pval == '"')
-							dres = strtod(pval + 1, &scan_length);
-						else
-							dres = strtod(pval, &scan_length);
-
-						if (isarray && *scan_length == '"')
-							scan_length++;
-
-						if (garbage_left(isarray, scan_length, compat))
-						{
-							ecpg_raise(lineno, ECPG_FLOAT_FORMAT,
-									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-							return (false);
-						}
-						pval = scan_length;
+						ecpg_raise(lineno, ECPG_FLOAT_FORMAT,
+								   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+						return (false);
 					}
-					else
-						dres = 0.0;
+					pval = scan_length;
 
 					switch (type)
 					{
@@ -349,37 +412,36 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 					break;
 
 				case ECPGt_bool:
-					if (pval)
+					if (pval[0] == 'f' && pval[1] == '\0')
 					{
-						if (pval[0] == 'f' && pval[1] == '\0')
-						{
-							if (offset == sizeof(char))
-								*((char *) (var + offset * act_tuple)) = false;
-							else if (offset == sizeof(int))
-								*((int *) (var + offset * act_tuple)) = false;
-							else
-								ecpg_raise(lineno, ECPG_CONVERT_BOOL,
-										   ECPG_SQLSTATE_DATATYPE_MISMATCH,
-										   NULL);
-							break;
-						}
-						else if (pval[0] == 't' && pval[1] == '\0')
-						{
-							if (offset == sizeof(char))
-								*((char *) (var + offset * act_tuple)) = true;
-							else if (offset == sizeof(int))
-								*((int *) (var + offset * act_tuple)) = true;
-							else
-								ecpg_raise(lineno, ECPG_CONVERT_BOOL,
-										   ECPG_SQLSTATE_DATATYPE_MISMATCH,
-										   NULL);
-							break;
-						}
-						else if (pval[0] == '\0' && PQgetisnull(results, act_tuple, act_field))
-						{
-							/* NULL is valid */
-							break;
-						}
+						if (offset == sizeof(char))
+							*((char *) (var + offset * act_tuple)) = false;
+						else if (offset == sizeof(int))
+							*((int *) (var + offset * act_tuple)) = false;
+						else
+							ecpg_raise(lineno, ECPG_CONVERT_BOOL,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH,
+									   NULL);
+						pval++;
+						break;
+					}
+					else if (pval[0] == 't' && pval[1] == '\0')
+					{
+						if (offset == sizeof(char))
+							*((char *) (var + offset * act_tuple)) = true;
+						else if (offset == sizeof(int))
+							*((int *) (var + offset * act_tuple)) = true;
+						else
+							ecpg_raise(lineno, ECPG_CONVERT_BOOL,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH,
+									   NULL);
+						pval++;
+						break;
+					}
+					else if (pval[0] == '\0' && PQgetisnull(results, act_tuple, act_field))
+					{
+						/* NULL is valid */
+						break;
 					}
 
 					ecpg_raise(lineno, ECPG_CONVERT_BOOL,
@@ -389,13 +451,28 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 
 				case ECPGt_char:
 				case ECPGt_unsigned_char:
-					if (pval)
+				case ECPGt_string:
 					{
+						char	   *str = (char *) (var + offset * act_tuple);
+
 						if (varcharsize == 0 || varcharsize > size)
-							strncpy((char *) ((long) var + offset * act_tuple), pval, size + 1);
+						{
+							strncpy(str, pval, size + 1);
+							/* do the rtrim() */
+							if (type == ECPGt_string)
+							{
+								char	   *last = str + size;
+
+								while (last > str && (*last == ' ' || *last == '\0'))
+								{
+									*last = '\0';
+									last--;
+								}
+							}
+						}
 						else
 						{
-							strncpy((char *) ((long) var + offset * act_tuple), pval, varcharsize);
+							strncpy(str, pval, varcharsize);
 
 							if (varcharsize < size)
 							{
@@ -414,12 +491,12 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 									case ECPGt_unsigned_long:
 										*((long *) (ind + ind_offset * act_tuple)) = size;
 										break;
-#ifdef HAVE_LONG_LONG_INT_64
+#ifdef HAVE_LONG_LONG_INT
 									case ECPGt_long_long:
 									case ECPGt_unsigned_long_long:
 										*((long long int *) (ind + ind_offset * act_tuple)) = size;
 										break;
-#endif   /* HAVE_LONG_LONG_INT_64 */
+#endif   /* HAVE_LONG_LONG_INT */
 									default:
 										break;
 								}
@@ -431,10 +508,9 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 					break;
 
 				case ECPGt_varchar:
-					if (pval)
 					{
 						struct ECPGgeneric_varchar *variable =
-						(struct ECPGgeneric_varchar *) ((long) var + offset * act_tuple);
+						(struct ECPGgeneric_varchar *) (var + offset * act_tuple);
 
 						variable->len = size;
 						if (varcharsize == 0)
@@ -450,22 +526,22 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 								{
 									case ECPGt_short:
 									case ECPGt_unsigned_short:
-										*((short *) (ind + ind_offset * act_tuple)) = variable->len;
+										*((short *) (ind + offset * act_tuple)) = variable->len;
 										break;
 									case ECPGt_int:
 									case ECPGt_unsigned_int:
-										*((int *) (ind + ind_offset * act_tuple)) = variable->len;
+										*((int *) (ind + offset * act_tuple)) = variable->len;
 										break;
 									case ECPGt_long:
 									case ECPGt_unsigned_long:
-										*((long *) (ind + ind_offset * act_tuple)) = variable->len;
+										*((long *) (ind + offset * act_tuple)) = variable->len;
 										break;
-#ifdef HAVE_LONG_LONG_INT_64
+#ifdef HAVE_LONG_LONG_INT
 									case ECPGt_long_long:
 									case ECPGt_unsigned_long_long:
 										*((long long int *) (ind + ind_offset * act_tuple)) = variable->len;
 										break;
-#endif   /* HAVE_LONG_LONG_INT_64 */
+#endif   /* HAVE_LONG_LONG_INT */
 									default:
 										break;
 								}
@@ -480,214 +556,199 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 
 				case ECPGt_decimal:
 				case ECPGt_numeric:
-					if (pval)
+					if (isarray && *pval == '"')
+						nres = PGTYPESnumeric_from_asc(pval + 1, &scan_length);
+					else
+						nres = PGTYPESnumeric_from_asc(pval, &scan_length);
+
+					/* did we get an error? */
+					if (nres == NULL)
 					{
-						if (isarray && *pval == '"')
-							nres = PGTYPESnumeric_from_asc(pval + 1, &scan_length);
-						else
-							nres = PGTYPESnumeric_from_asc(pval, &scan_length);
+						ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
+								 lineno, pval ? pval : "", errno);
 
-						/* did we get an error? */
-						if (nres == NULL)
+						if (INFORMIX_MODE(compat))
 						{
-							ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
-									 lineno, pval ? pval : "", errno);
-
-							if (INFORMIX_MODE(compat))
-							{
-								/*
-								 * Informix wants its own NULL value here
-								 * instead of an error
-								 */
-								nres = PGTYPESnumeric_new();
-								if (nres)
-									ECPGset_noind_null(ECPGt_numeric, nres);
-								else
-								{
-									ecpg_raise(lineno, ECPG_OUT_OF_MEMORY,
-									 ECPG_SQLSTATE_ECPG_OUT_OF_MEMORY, NULL);
-									return (false);
-								}
-							}
+							/*
+							 * Informix wants its own NULL value here instead
+							 * of an error
+							 */
+							nres = PGTYPESnumeric_new();
+							if (nres)
+								ECPGset_noind_null(ECPGt_numeric, nres);
 							else
 							{
-								ecpg_raise(lineno, ECPG_NUMERIC_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+								ecpg_raise(lineno, ECPG_OUT_OF_MEMORY,
+									 ECPG_SQLSTATE_ECPG_OUT_OF_MEMORY, NULL);
 								return (false);
 							}
 						}
 						else
 						{
-							if (isarray && *scan_length == '"')
-								scan_length++;
-
-							if (garbage_left(isarray, scan_length, compat))
-							{
-								free(nres);
-								ecpg_raise(lineno, ECPG_NUMERIC_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-								return (false);
-							}
+							ecpg_raise(lineno, ECPG_NUMERIC_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
 						}
-						pval = scan_length;
 					}
 					else
-						nres = PGTYPESnumeric_from_asc("0.0", &scan_length);
+					{
+						if (isarray && *scan_length == '"')
+							scan_length++;
+
+						if (garbage_left(isarray, scan_length, compat))
+						{
+							free(nres);
+							ecpg_raise(lineno, ECPG_NUMERIC_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
+						}
+					}
+					pval = scan_length;
 
 					if (type == ECPGt_numeric)
 						PGTYPESnumeric_copy(nres, (numeric *) (var + offset * act_tuple));
 					else
 						PGTYPESnumeric_to_decimal(nres, (decimal *) (var + offset * act_tuple));
 
-					free(nres);
+					PGTYPESnumeric_free(nres);
 					break;
 
 				case ECPGt_interval:
-					if (pval)
+					if (isarray && *pval == '"')
+						ires = PGTYPESinterval_from_asc(pval + 1, &scan_length);
+					else
+						ires = PGTYPESinterval_from_asc(pval, &scan_length);
+
+					/* did we get an error? */
+					if (ires == NULL)
 					{
-						if (isarray && *pval == '"')
-							ires = PGTYPESinterval_from_asc(pval + 1, &scan_length);
-						else
-							ires = PGTYPESinterval_from_asc(pval, &scan_length);
+						ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
+								 lineno, pval ? pval : "", errno);
 
-						/* did we get an error? */
-						if (ires == NULL)
+						if (INFORMIX_MODE(compat))
 						{
-							ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
-									 lineno, pval ? pval : "", errno);
-
-							if (INFORMIX_MODE(compat))
-							{
-								/*
-								 * Informix wants its own NULL value here
-								 * instead of an error
-								 */
-								ires = (interval *) ecpg_alloc(sizeof(interval), lineno);
-								if (!ires)
-									return (false);
-
-								ECPGset_noind_null(ECPGt_interval, ires);
-							}
-							else
-							{
-								ecpg_raise(lineno, ECPG_INTERVAL_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							/*
+							 * Informix wants its own NULL value here instead
+							 * of an error
+							 */
+							ires = (interval *) ecpg_alloc(sizeof(interval), lineno);
+							if (!ires)
 								return (false);
-							}
+
+							ECPGset_noind_null(ECPGt_interval, ires);
 						}
 						else
 						{
-							if (isarray && *scan_length == '"')
-								scan_length++;
-
-							if (garbage_left(isarray, scan_length, compat))
-							{
-								free(ires);
-								ecpg_raise(lineno, ECPG_INTERVAL_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-								return (false);
-							}
+							ecpg_raise(lineno, ECPG_INTERVAL_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
 						}
-						pval = scan_length;
 					}
 					else
-						ires = PGTYPESinterval_from_asc("0 seconds", NULL);
+					{
+						if (isarray && *scan_length == '"')
+							scan_length++;
+
+						if (garbage_left(isarray, scan_length, compat))
+						{
+							free(ires);
+							ecpg_raise(lineno, ECPG_INTERVAL_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
+						}
+					}
+					pval = scan_length;
 
 					PGTYPESinterval_copy(ires, (interval *) (var + offset * act_tuple));
 					free(ires);
 					break;
+
 				case ECPGt_date:
-					if (pval)
+					if (isarray && *pval == '"')
+						ddres = PGTYPESdate_from_asc(pval + 1, &scan_length);
+					else
+						ddres = PGTYPESdate_from_asc(pval, &scan_length);
+
+					/* did we get an error? */
+					if (errno != 0)
 					{
-						if (isarray && *pval == '"')
-							ddres = PGTYPESdate_from_asc(pval + 1, &scan_length);
-						else
-							ddres = PGTYPESdate_from_asc(pval, &scan_length);
+						ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
+								 lineno, pval ? pval : "", errno);
 
-						/* did we get an error? */
-						if (errno != 0)
+						if (INFORMIX_MODE(compat))
 						{
-							ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
-									 lineno, pval ? pval : "", errno);
-
-							if (INFORMIX_MODE(compat))
-							{
-								/*
-								 * Informix wants its own NULL value here
-								 * instead of an error
-								 */
-								ECPGset_noind_null(ECPGt_date, &ddres);
-							}
-							else
-							{
-								ecpg_raise(lineno, ECPG_DATE_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-								return (false);
-							}
+							/*
+							 * Informix wants its own NULL value here instead
+							 * of an error
+							 */
+							ECPGset_noind_null(ECPGt_date, &ddres);
 						}
 						else
 						{
-							if (isarray && *scan_length == '"')
-								scan_length++;
-
-							if (garbage_left(isarray, scan_length, compat))
-							{
-								ecpg_raise(lineno, ECPG_DATE_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-								return (false);
-							}
+							ecpg_raise(lineno, ECPG_DATE_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
 						}
-
-						*((date *) (var + offset * act_tuple)) = ddres;
-						pval = scan_length;
 					}
+					else
+					{
+						if (isarray && *scan_length == '"')
+							scan_length++;
+
+						if (garbage_left(isarray, scan_length, compat))
+						{
+							ecpg_raise(lineno, ECPG_DATE_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
+						}
+					}
+
+					*((date *) (var + offset * act_tuple)) = ddres;
+					pval = scan_length;
 					break;
 
 				case ECPGt_timestamp:
-					if (pval)
+					if (isarray && *pval == '"')
+						tres = PGTYPEStimestamp_from_asc(pval + 1, &scan_length);
+					else
+						tres = PGTYPEStimestamp_from_asc(pval, &scan_length);
+
+					/* did we get an error? */
+					if (errno != 0)
 					{
-						if (isarray && *pval == '"')
-							tres = PGTYPEStimestamp_from_asc(pval + 1, &scan_length);
-						else
-							tres = PGTYPEStimestamp_from_asc(pval, &scan_length);
+						ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
+								 lineno, pval ? pval : "", errno);
 
-						/* did we get an error? */
-						if (errno != 0)
+						if (INFORMIX_MODE(compat))
 						{
-							ecpg_log("ecpg_get_data on line %d: RESULT %s; errno %d\n",
-									 lineno, pval ? pval : "", errno);
-
-							if (INFORMIX_MODE(compat))
-							{
-								/*
-								 * Informix wants its own NULL value here
-								 * instead of an error
-								 */
-								ECPGset_noind_null(ECPGt_timestamp, &tres);
-							}
-							else
-							{
-								ecpg_raise(lineno, ECPG_TIMESTAMP_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-								return (false);
-							}
+							/*
+							 * Informix wants its own NULL value here instead
+							 * of an error
+							 */
+							ECPGset_noind_null(ECPGt_timestamp, &tres);
 						}
 						else
 						{
-							if (isarray && *scan_length == '"')
-								scan_length++;
-
-							if (garbage_left(isarray, scan_length, compat))
-							{
-								ecpg_raise(lineno, ECPG_TIMESTAMP_FORMAT,
-									  ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
-								return (false);
-							}
+							ecpg_raise(lineno, ECPG_TIMESTAMP_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
 						}
-
-						*((timestamp *) (var + offset * act_tuple)) = tres;
-						pval = scan_length;
 					}
+					else
+					{
+						if (isarray && *scan_length == '"')
+							scan_length++;
+
+						if (garbage_left(isarray, scan_length, compat))
+						{
+							ecpg_raise(lineno, ECPG_TIMESTAMP_FORMAT,
+									   ECPG_SQLSTATE_DATATYPE_MISMATCH, pval);
+							return (false);
+						}
+					}
+
+					*((timestamp *) (var + offset * act_tuple)) = tres;
+					pval = scan_length;
 					break;
 
 				default:
@@ -697,7 +758,7 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 					return (false);
 					break;
 			}
-			if (isarray == ECPG_ARRAY_ARRAY)
+			if (ECPG_IS_ARRAY(isarray))
 			{
 				bool		string = false;
 
@@ -705,30 +766,20 @@ ecpg_get_data(const PGresult *results, int act_tuple, int act_field, int lineno,
 				++act_tuple;
 
 				/* set pval to the next entry */
-				for (; string || (*pval != ',' && *pval != '}' && *pval != '\0'); ++pval)
+
+				/*
+				 * *pval != '\0' should not be needed, but is used as a safety
+				 * guard
+				 */
+				for (; *pval != '\0' && (string || (!array_delimiter(isarray, *pval) && !array_boundary(isarray, *pval))); ++pval)
 					if (*pval == '"')
 						string = string ? false : true;
 
-				if (*pval == ',')
-					++pval;
-			}
-			else if (isarray == ECPG_ARRAY_VECTOR)
-			{
-				bool		string = false;
-
-				/* set array to next entry */
-				++act_tuple;
-
-				/* set pval to the next entry */
-				for (; string || (*pval != ' ' && *pval != '\0'); ++pval)
-					if (*pval == '"')
-						string = string ? false : true;
-
-				if (*pval == ' ')
+				if (array_delimiter(isarray, *pval))
 					++pval;
 			}
 		}
-	} while (*pval != '\0' && ((isarray == ECPG_ARRAY_ARRAY && *pval != '}') || isarray == ECPG_ARRAY_VECTOR));
+	} while (*pval != '\0' && !array_boundary(isarray, *pval));
 
 	return (true);
 }
